//
#include <linux/device.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/cdev.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

#include "mod0.h"


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Igor D: 19.08.2021");
MODULE_DESCRIPTION("mod1 testing.");
MODULE_VERSION("1.01");
MODULE_SOFTDEP("pre: mod0");
//MODULE_DEPENDS(mod0, 1);

//EXPORT_NO_SYMBOLS;

#define DEVICE_NAME 	"test_mod1"
#define EXAMPLE_MSG 	"mod1 loaded!\n"
#define MSG_BUFFER_LEN 	32
//#define SZ_VAL		10

static dev_t first;		// Global variable for the first device number
static struct cdev c_dev;	// Global variable for the character device structure
static struct class *cl;	// Global variable for the device class
//
//static int strreverse(char *instr, char *outstr, size_t len);
//
static int major_num;
static int device_open_count = 0;
static char msg_buffer[MSG_BUFFER_LEN];
static char *msg_ptr;
static int cnt_rd = 0;

/* Prototypes for device functions */
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);

/* This structure points to all of the device functions */
static struct file_operations file_ops = {
	.read = device_read,
	.write = device_write,
	.open = device_open,
	.release = device_release
	//
	
};


/* When a process reads from our device, this gets called. */
static ssize_t device_read(struct file *flip, char *buffer, size_t len, loff_t *offset) {
	int bytes_read = 0;
 	/* If we’re at the end, loop back to the beginning */
 	if (*msg_ptr == 0) {
		if(cnt_rd > 0) {
 			msg_ptr = msg_buffer;
			cnt_rd = 0;
		} else {
			return 0;
		}
 	}
 	/* Put data in the buffer */
 	while (len && *msg_ptr) {
 		/* Buffer is in user data, not kernel, so you can’t just reference
	 	* with a pointer. The function put_user handles this for us */
 		put_user(*(msg_ptr++), buffer++);
 		len--;
 		bytes_read++;
 	}
 	return bytes_read;
}

/* Called when a process tries to write to our device */
static ssize_t device_write(struct file *flip, const char *buffer, size_t len, loff_t *offset) {
	//char outstr[MSG_BUFFER_LEN] = {0};
	if(len < MSG_BUFFER_LEN) {
		memset(instr, 0, MSG_BUFFER_LEN);
		memset(outstr, 0, MSG_BUFFER_LEN);
		memcpy(instr, buffer, (int)(len));
		printk(KERN_INFO "from /dev/test_mod1:%s len=%i\n", outstr, (int)(len));
		//
		mod0_strreverse(instr, len);
		len = mod0_outstr(outstr);
		//
		memcpy(msg_buffer, outstr, (int)(len));
		printk(KERN_INFO "from mod0:%s len=%i\n", outstr, (int)(len));
		cnt_rd = len;
		//
		return len;
	} else 
		return -EINVAL;
}

/* Called when a process opens our device */
static int device_open(struct inode *inode, struct file *file) {
	/* If device is open, return busy */
	if (device_open_count) {
		return -EBUSY;
	}
	device_open_count++;
	try_module_get(THIS_MODULE);
	return 0;
}

/* Called when a process closes our device */
static int device_release(struct inode *inode, struct file *file) {
	/* Decrement the open counter and usage count. Without this, the module would not unload. */
	device_open_count--;
	module_put(THIS_MODULE);
	return 0;
}


static int __init mod1_init(void) {

	printk(KERN_INFO "Namaskar: ofcd registered\n");
	if (alloc_chrdev_region(&first, 0, 1, DEVICE_NAME) < 0)
	{
		return -1;
	}
	if ((cl = class_create(THIS_MODULE, DEVICE_NAME)) == NULL)
	{
		unregister_chrdev_region(first, 1);
		return -1;
	}
	if (device_create(cl, NULL, first, NULL, DEVICE_NAME) == NULL)
	{
		class_destroy(cl);
		unregister_chrdev_region(first, 1);
		return -1;
	}
	//cdev_init(&c_dev, &pugs_fops);
	cdev_init(&c_dev, &file_ops);
        if (cdev_add(&c_dev, first, 1) == -1)
	{
		device_destroy(cl, first);
		class_destroy(cl);
		unregister_chrdev_region(first, 1);
		return -1;
	}
 	/* Fill buffer with our message */
 	strncpy(msg_buffer, EXAMPLE_MSG, MSG_BUFFER_LEN);
	/* Set the msg_ptr to the buffer */
	msg_ptr = msg_buffer;
	/* Try to register character device */
	major_num = register_chrdev(0, DEVICE_NAME, &file_ops);
	if (major_num < 0) {
	printk(KERN_ALERT "Could not register device: %d\n", major_num);
	return major_num;
	} else {
		printk(KERN_INFO "mod1 module loaded!\nwith device major number %d\n", major_num);
		return 0;
	}
}

static void __exit mod1_exit(void) {
	cdev_del(&c_dev);
	device_destroy(cl, first);
	class_destroy(cl);
	/* Remember — we have to clean up after ourselves. Unregister the character device. */
	unregister_chrdev(major_num, DEVICE_NAME);
	printk(KERN_INFO "mod1 unloaded!\n");
}

/* Register module functions */
module_init(mod1_init);
module_exit(mod1_exit);
